@@ -6,6 +6,7 @@ import (
 	"context"
 	"encoding/json"
 	"errors"
+	"expvar"
 	"fmt"
 	"sort"
 	"sync"
@@ -15,7 +16,6 @@ import (
 
 	"github.com/creachadair/jrpc2"
 	"github.com/creachadair/jrpc2/channel"
-	"github.com/creachadair/jrpc2/code"
 	"github.com/creachadair/jrpc2/handler"
 	"github.com/creachadair/jrpc2/server"
 	"github.com/fortytw2/leaktest"
@@ -24,7 +24,7 @@ import (
 
 // Static type assertions.
 var (
-	_ code.ErrCoder = (*jrpc2.Error)(nil)
+	_ jrpc2.ErrCoder = (*jrpc2.Error)(nil)
 )
 
 var testOK = handler.New(func(ctx context.Context) (string, error) {
@@ -41,12 +41,13 @@ var testService = handler.Map{
 	"Nil":  handler.New(methodNil),
 	"Ctx":  handler.New(methodCtx),
 	"Ping": handler.New(methodPing),
+	"Echo": handler.New(methodStruct),
 }
 
 type dummy struct{}
 
 // Add is a request-based method.
-func (dummy) Add(_ context.Context, req *jrpc2.Request) (interface{}, error) {
+func (dummy) Add(_ context.Context, req *jrpc2.Request) (any, error) {
 	if req.IsNotification() {
 		return nil, errors.New("ignoring notification")
 	}
@@ -69,7 +70,7 @@ func (dummy) Mul(_ context.Context, req struct{ X, Y int }) (int, error) {
 // Max takes a slice of arguments.
 func (dummy) Max(_ context.Context, vs []int) (int, error) {
 	if len(vs) == 0 {
-		return 0, jrpc2.Errorf(code.InvalidParams, "cannot compute max of no elements")
+		return 0, jrpc2.Errorf(jrpc2.InvalidParams, "cannot compute max of no elements")
 	}
 	max := vs[0]
 	for _, v := range vs[1:] {
@@ -99,9 +100,30 @@ func methodPing(ctx context.Context, req *jrpc2.Request) error {
 	return nil
 }
 
+type Anon struct{ OK string }
+
+// structArgs has 4 valid fields.
+type structArgs struct {
+	A string `json:"alpha"` // explicitly named
+	B bool   `json:"-"`     // explicitly skipped
+
+	//lint:ignore U1000 Verify that unexported fields are skipped.
+	skip bool
+
+	Anon     // unnamed anonymous, skipped for positional
+	C    int // default name "c"
+	D    int `json:"delta"` // assigned name "delta"
+}
+
+// methodStruct has a struct argument that can be called with either an object
+// or an array. It returns the value of arg.D.
+func methodStruct(ctx context.Context, arg *structArgs) int {
+	return arg.D
+}
+
 var callTests = []struct {
 	method string
-	params interface{}
+	params any
 	want   int
 }{
 	{"Test.Add", []int{}, 0},
@@ -112,6 +134,8 @@ var callTests = []struct {
 	{"Test.Ctx", nil, 1},
 	{"Test.Nil", nil, 42},
 	{"Test.Nil", json.RawMessage("null"), 42},
+	{"Test.Echo", []any{"foo", 4, 17}, 17},
+	{"Test.Echo", map[string]any{"delta": 144, "ok": "yes"}, 144},
 }
 
 func TestServerInfo_methodNames(t *testing.T) {
@@ -125,7 +149,7 @@ func TestServerInfo_methodNames(t *testing.T) {
 
 	// Verify that the assigner got the names it was supposed to.
 	got, want := s.ServerInfo().Methods, []string{
-		"Test.Add", "Test.Ctx", "Test.Max", "Test.Mul", "Test.Nil", "Test.Ping",
+		"Test.Add", "Test.Ctx", "Test.Echo", "Test.Max", "Test.Mul", "Test.Nil", "Test.Ping",
 	}
 	if diff := cmp.Diff(want, got); diff != "" {
 		t.Errorf("Wrong method names: (-want, +got)\n%s", diff)
@@ -381,8 +405,8 @@ func TestServer_stopCancelsHandlers(t *testing.T) {
 	case <-time.After(30 * time.Second):
 		t.Error("Timed out waiting for service handler to fail")
 	case err := <-stopped:
-		if ec := code.FromError(err); ec != code.Cancelled {
-			t.Errorf("Client error: got %v (%v), wanted code %v", err, ec, code.Cancelled)
+		if ec := jrpc2.ErrorCode(err); ec != jrpc2.Cancelled {
+			t.Errorf("Client error: got %v (%v), wanted code %v", err, ec, jrpc2.Cancelled)
 		}
 	}
 }
@@ -431,9 +455,9 @@ func TestServer_CancelRequest(t *testing.T) {
 	}
 
 	err := <-errc
-	got := code.FromError(err)
-	if got != code.Cancelled {
-		t.Errorf("Stall: got %v (%v), want %v", err, got, code.Cancelled)
+	got := jrpc2.ErrorCode(err)
+	if got != jrpc2.Cancelled {
+		t.Errorf("Stall: got %v (%v), want %v", err, got, jrpc2.Cancelled)
 	} else {
 		t.Logf("Cancellation succeeded, got expected error: %v", err)
 	}
@@ -455,7 +479,7 @@ func TestError_withData(t *testing.T) {
 			return false, jrpc2.ServerFromContext(ctx).Notify(ctx, "PushBack", nil)
 		}),
 		"Code": handler.New(func(ctx context.Context) error {
-			return code.Code(12345).Err()
+			return jrpc2.Code(12345).Err()
 		}),
 	}, &server.LocalOptions{
 		Client: &jrpc2.ClientOptions{
@@ -499,7 +523,7 @@ func TestClient_badCallParams(t *testing.T) {
 	defer leaktest.Check(t)()
 
 	loc := server.NewLocal(handler.Map{
-		"Test": handler.New(func(_ context.Context, v interface{}) error {
+		"Test": handler.New(func(_ context.Context, v any) error {
 			return jrpc2.Errorf(129, "this should not be reached")
 		}),
 	}, nil)
@@ -508,7 +532,7 @@ func TestClient_badCallParams(t *testing.T) {
 	rsp, err := loc.Client.Call(context.Background(), "Test", "bogus")
 	if err == nil {
 		t.Errorf("Call(Test): got %+v, wanted error", rsp)
-	} else if got, want := code.FromError(err), code.InvalidRequest; got != want {
+	} else if got, want := jrpc2.ErrorCode(err), jrpc2.InvalidRequest; got != want {
 		t.Errorf("Call(Test): got code %v, want %v", got, want)
 	}
 }
@@ -517,27 +541,13 @@ func TestClient_badCallParams(t *testing.T) {
 func TestServer_serverInfoMetrics(t *testing.T) {
 	defer leaktest.Check(t)()
 
+	const metricValue = "if red were green, green would be my favourite colour"
+	customMetric := new(expvar.String)
+	customMetric.Set(metricValue)
+
 	loc := server.NewLocal(handler.Map{
 		"Metricize": handler.New(func(ctx context.Context) (bool, error) {
-			m := jrpc2.ServerFromContext(ctx).Metrics()
-			if m == nil {
-				t.Error("Request context does not contain a metrics writer")
-				return false, nil
-			}
-			m.Count("counters-written", 1)
-			m.Count("counters-written", 2)
-
-			// Max value trackers are not accumulative.
-			m.SetMaxValue("max-metric-value", 1)
-			m.SetMaxValue("max-metric-value", 5)
-			m.SetMaxValue("max-metric-value", 3)
-			m.SetMaxValue("max-metric-value", -30337)
-
-			// Counters are accumulative, and negative deltas subtract.
-			m.Count("zero-sum", 0)
-			m.Count("zero-sum", 15)
-			m.Count("zero-sum", -16)
-			m.Count("zero-sum", 1)
+			jrpc2.ServerMetrics().Set("custom-metric", customMetric)
 			return true, nil
 		}),
 	}, nil)
@@ -547,36 +557,23 @@ func TestServer_serverInfoMetrics(t *testing.T) {
 	if _, err := c.Call(ctx, "Metricize", nil); err != nil {
 		t.Fatalf("Call(Metricize) failed: %v", err)
 	}
-	if got := s.ServerInfo().Counter["rpc.serversActive"]; got != 1 {
-		t.Errorf("Metric rpc.serversActive: got %d, want 1", got)
+	active, ok := jrpc2.ServerMetrics().Get("servers_active").(*expvar.Int)
+	if !ok {
+		t.Fatal("Metric servers_active not defined")
+	} else if active.Value() != 2 { // 1 for this test, 1 for the examples
+		t.Errorf("Metric servers_active: got %d, want 1", active.Value())
 	}
 	loc.Close()
 
-	info := s.ServerInfo()
-	tests := []struct {
-		input map[string]int64
-		name  string
-		want  int64 // use < 0 to test for existence only
-	}{
-		{info.Counter, "rpc.requests", 1},
-		{info.Counter, "counters-written", 3},
-		{info.Counter, "zero-sum", 0},
-		{info.Counter, "rpc.bytesRead", -1},
-		{info.Counter, "rpc.bytesWritten", -1},
-		{info.Counter, "rpc.serversActive", 0},
-		{info.MaxValue, "max-metric-value", 5},
-		{info.MaxValue, "rpc.bytesRead", -1},
-		{info.MaxValue, "rpc.bytesWritten", -1},
-	}
-	for _, test := range tests {
-		got, ok := test.input[test.name]
-		if !ok {
-			t.Errorf("Metric %q is not defined, but was expected", test.name)
-			continue
-		}
-		if test.want >= 0 && got != test.want {
-			t.Errorf("Wrong value for metric %q: got %d, want %d", test.name, got, test.want)
-		}
+	// When populated into a ServerInfo value, metrics should be converted to
+	// JSON values in a sensible way.
+	metric := s.ServerInfo().Metrics["custom-metric"]
+	if metric == nil {
+		t.Fatal("Custom metric not found")
+	} else if got, ok := metric.(json.RawMessage); !ok {
+		t.Fatalf("Wrong metric type: got %T, want json.RawMessage", metric)
+	} else if s := string(got); s != `"`+metricValue+`"` {
+		t.Fatalf("Wrong metric value: got %q, want %q", s, metricValue)
 	}
 }
 
@@ -589,7 +586,7 @@ func TestServer_nonLibraryClient(t *testing.T) {
 	srv, cli := channel.Direct()
 	s := jrpc2.NewServer(handler.Map{
 		"X": testOK,
-		"Y": handler.New(func(context.Context) (interface{}, error) {
+		"Y": handler.New(func(context.Context) (any, error) {
 			return nil, nil
 		}),
 	}, nil).Start(srv)
@@ -805,7 +802,7 @@ func TestServer_Callback(t *testing.T) {
 	}, &server.LocalOptions{
 		Server: &jrpc2.ServerOptions{AllowPush: true},
 		Client: &jrpc2.ClientOptions{
-			OnCallback: func(ctx context.Context, req *jrpc2.Request) (interface{}, error) {
+			OnCallback: func(ctx context.Context, req *jrpc2.Request) (any, error) {
 				t.Logf("OnCallback invoked for method %q", req.Method())
 				switch req.Method() {
 				case "succeed":
@@ -871,7 +868,7 @@ func TestClient_onCancelHook(t *testing.T) {
 		"computerSaysNo": handler.New(func(ctx context.Context, ids []string) error {
 			defer close(hooked)
 			if req := jrpc2.InboundRequest(ctx); !req.IsNotification() {
-				return jrpc2.Errorf(code.MethodNotFound, "no such method %q", req.Method())
+				return jrpc2.Errorf(jrpc2.MethodNotFound, "no such method %q", req.Method())
 			}
 			srv := jrpc2.ServerFromContext(ctx)
 			for _, id := range ids {
@@ -994,7 +991,7 @@ func TestClient_concurrentCallbacks(t *testing.T) {
 	}, &server.LocalOptions{
 		Server: &jrpc2.ServerOptions{AllowPush: true},
 		Client: &jrpc2.ClientOptions{
-			OnCallback: handler.Func(func(ctx context.Context, req *jrpc2.Request) (interface{}, error) {
+			OnCallback: func(ctx context.Context, req *jrpc2.Request) (any, error) {
 				// A trivial callback that reports its method name.
 				// The name is used to select which invocation we are serving.
 				switch req.Method() {
@@ -1007,7 +1004,7 @@ func TestClient_concurrentCallbacks(t *testing.T) {
 				}
 				<-release
 				return req.Method(), nil
-			}),
+			},
 		},
 	})
 	defer loc.Close()
@@ -1071,8 +1068,8 @@ func TestHandler_noParams(t *testing.T) {
 	var rsp string
 	if err := loc.Client.CallResult(context.Background(), "Test", []int{1, 2, 3}, &rsp); err == nil {
 		t.Errorf("Call(Test): got %q, wanted error", rsp)
-	} else if ec := code.FromError(err); ec != code.InvalidParams {
-		t.Errorf("Call(Test): got code %v, wanted %v", ec, code.InvalidParams)
+	} else if ec := jrpc2.ErrorCode(err); ec != jrpc2.InvalidParams {
+		t.Errorf("Call(Test): got code %v, wanted %v", ec, jrpc2.InvalidParams)
 	}
 }
 
@@ -1083,9 +1080,9 @@ func TestRPCServerInfo(t *testing.T) {
 	loc := server.NewLocal(handler.Map{"Test": testOK}, nil)
 	defer loc.Close()
 
-	si, err := jrpc2.RPCServerInfo(context.Background(), loc.Client)
-	if err != nil {
-		t.Errorf("RPCServerInfo failed: %v", err)
+	var si jrpc2.ServerInfo
+	if err := loc.Client.CallResult(context.Background(), "rpc.serverInfo", nil, &si); err != nil {
+		t.Errorf("rpc.serverInfo call failed: %v", err)
 	}
 	{
 		got, want := si.Methods, []string{"Test"}
@@ -1234,18 +1231,18 @@ func TestRequest_strictFields(t *testing.T) {
 
 	tests := []struct {
 		method string
-		params interface{}
-		code   code.Code
+		params any
+		code   jrpc2.Code
 		want   string
 	}{
-		{"Strict", handler.Obj{"alpha": "aiuto"}, code.NoError, "aiuto"},
-		{"Strict", handler.Obj{"alpha": "selva me", "charlie": true}, code.NoError, "selva me"},
-		{"Strict", handler.Obj{"alpha": "OK", "nonesuch": true}, code.InvalidParams, ""},
-		{"Normal", handler.Obj{"alpha": "OK", "nonesuch": true}, code.NoError, "OK"},
+		{"Strict", handler.Obj{"alpha": "aiuto"}, jrpc2.NoError, "aiuto"},
+		{"Strict", handler.Obj{"alpha": "selva me", "charlie": true}, jrpc2.NoError, "selva me"},
+		{"Strict", handler.Obj{"alpha": "OK", "nonesuch": true}, jrpc2.InvalidParams, ""},
+		{"Normal", handler.Obj{"alpha": "OK", "nonesuch": true}, jrpc2.NoError, "OK"},
 	}
 	for _, test := range tests {
 		name := test.method + "/"
-		if test.code == code.NoError {
+		if test.code == jrpc2.NoError {
 			name += "OK"
 		} else {
 			name += test.code.String()
@@ -1253,10 +1250,10 @@ func TestRequest_strictFields(t *testing.T) {
 		t.Run(name, func(t *testing.T) {
 			var res string
 			err := loc.Client.CallResult(ctx, test.method, test.params, &res)
-			if err == nil && test.code != code.NoError {
+			if err == nil && test.code != jrpc2.NoError {
 				t.Errorf("CallResult: got %+v, want error code %v", res, test.code)
 			} else if err != nil {
-				if c := code.FromError(err); c != test.code {
+				if c := jrpc2.ErrorCode(err); c != test.code {
 					t.Errorf("CallResult: got error %v, wanted code %v", err, test.code)
 				}
 			} else if res != test.want {
@@ -1351,3 +1348,32 @@ func TestServer_newContext(t *testing.T) {
 		t.Errorf("Call failed: %v", err)
 	}
 }
+
+func TestClient_IsStopped(t *testing.T) {
+	clich := make(chan *jrpc2.Client, 1)
+	loc := server.NewLocal(handler.Map{}, &server.LocalOptions{
+		Client: &jrpc2.ClientOptions{
+			OnStop: func(cli *jrpc2.Client, _ error) { clich <- cli },
+		},
+	})
+
+	// A client should not be initially stopped.
+	if loc.Client.IsStopped() {
+		t.Error("IsStopped is true, want false")
+	}
+
+	// Closing the client should trigger its callback and leave it stopped.
+	loc.Client.Close()
+	loc.Server.Wait() // ensure callbacks have settled
+	close(clich)      // unblock
+
+	// A closed client should appear stopped.
+	if !loc.Client.IsStopped() {
+		t.Error("IsStopped is false, want true")
+	}
+
+	// The stop hook should have populated the same client value.
+	if cli := <-clich; cli != loc.Client {
+		t.Errorf("OnStop did not work: got %v, want %v", cli, loc.Client)
+	}
+}
